/*
** Module   : AstManager.java
** Abstract : provides services for creating, deleting, editing and using ASTs
**
** Copyright (c) 2009-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20090430   @42198 Created initial version. This singleton class
**                           provides services for creating, deleting, editing
**                           and using ASTs. These services are actually
**                           implemented inside a plugin that allows the
**                           services to be completely customized and replaced.
** 002 GES 20090518   @42368 Import change.
** 003 CA  20131023          Added some helpers related to the in-memory registry support.
*/

package com.goldencode.ast;

/**
 * This singleton class provides services for creating, deleting, editing
 * and using Abstract Syntax Trees (ASTs). These services are actually
 * implemented inside a plugin that allows the services to be completely
 * customized and replaced.
 * <p>
 * To create the singleton instance, the backing plugin of type
 * {@link AstManagerPlugin} must be installed via {@link #initialize}.
 * <p>
 * Maintains a project-specific bidirectional indexing facility for mapping
 * ASTs to project-unique IDs.  Given a 64-bit unique ID, one can lookup the
 * specific AST name that is associated with that ID. Likewise, given an AST
 * name, one can obtain the 64-bit unique ID.  In addition, this class tracks
 * the next available 64-bit node ID for each tree.
 * <p>
 * A <code>long</code> is used to store and represent the 64-bit IDs.  These 
 * 64-bit IDs are 'partitioned' into an upper DWORD (32-bit unsigned value) and
 * a lower DWORD. The upper DWORD is associated with the name that is the source
 * of a given AST.  Each tree name in a project will have a unique upper DWORD
 * that is assigned when the tree is added to the registry using the
 * {@link #addTree} method.  All AST nodes in the tree created from/for that
 * tree will share the same upper DWORD.  This upper DWORD is called the
 * 'tree portion' of the ID.  A 'tree ID' is an ID with the lower DWORD set
 * to 0x00000000 and the upper DWORD set to the tree portion of the specific
 * tree being referenced.  The lower DWORD holds a value that is unique to
 * each AST node in that tree.  This lower DWORD is called the 'node portion'
 * of the ID.  When a node portion is combined with the tree portion of an
 * ID, the result is a 64-bit ID that uniquely identifies a specific AST node
 * in a specific tree.  For this reason, such IDs can be used as persistent 
 * references between ASTs in different trees or even between AST nodes in
 * the same tree.
 * <p>
 * An ID of 0xFFFFFFFFFFFFFFFF is invalid and is the equivalent to a
 * <code>null</code> ID.  Similarly, a tree ID (which by definition has a
 * a lower DWORD of 0x00000000) can never refer to a specific node and only
 * refers to a tree (to the entire tree rather than any given node of that
 * tree).
 * <p>
 * This class provides the assignment facility for obtaining unique tree IDs
 * as well as unique node IDs (see {@link #getNextNodeId}).
 * <p>
 * <b>Name lookups are done on a case sensitive basis.</b>
 * <p>
 * To facilitate multi-threaded applications, a tree level locking facility
 * is provided. {@link #lockTree} and {@link #unlockTree} provide a mechanism
 * to obtain and release this lock respectively. This locking has nothing to
 * do with file system locks nor is there any non-Java locking backing this
 * lock. Rather, this lock is a cooperative agreement between all callers
 * that no use of the given tree will be made except when the caller holds
 * this exclusive lock to the tree.
 */
public class AstManager
{
   /** An invalid AST ID. */
   public static final long INVALID_ID = 0xFFFFFFFFFFFFFFFFL;
   
   /** Singleton instance of this class. */
   private static AstManager mgr = null;
   
   /** Core worker that provides the services implementation. */
   private AstManagerPlugin plugin = null;
   
   /**
    * Private constructor to enforce the singleton model.
    *
    * @param    plugin
    *           Core worker that provides the services implementation.
    *
    * @throws   AstException
    *           If the plugin instance provided is <code>null</code>.
    */
   private AstManager(AstManagerPlugin plugin)
   {
      this.plugin = plugin;
      
      if (plugin == null)
      {
         throw new AstException("AstManager plugin cannot be null!");
      }
   }
   
   /**
    * Instantiate the singleton instance (or return the already existing
    * instance). If the singleton instance must be created, then it will have
    * the given plugin as the core worker. Otherwise, whatever plugin was
    * specified at instantiation will be retained and this given plugin will
    * be ignored.
    *
    * @param    plugin
    *           Core worker that provides the services implementation. Ignored
    *           if the singleton instance already exists.
    *
    * @return   The singleton instance of this class.
    */
   public static synchronized AstManager initialize(AstManagerPlugin plugin)
   {
      if (mgr == null)
      {
         mgr = new AstManager(plugin);
      }
      
      return mgr;
   }
   
   /**
    * Obtain the singleton instance of this class.
    *
    * @return   The singleton instance of this class.
    *
    * @throws   AstException
    *           If the singleton instance has not yet been initialized.
    */
   public static synchronized AstManager get()
   {
      if (mgr == null)
      {
         throw new AstException("AstManager is not yet initialized!");
      }
      
      return mgr;
   }
   
   /**
    * Check if the plugin is of the specified type.
    * 
    * @param    type
    *           The expected type.
    *           
    * @return   <code>true</code> if the {@link #plugin} has the same or is a sub-type of the
    *           specified type.
    */
   public static boolean isPluginOfType(Class<?> type)
   {
      return type.isAssignableFrom(get().plugin.getClass());
   }
   
   /**
    * Lookup the name associated with the tree specified by this ID. The ID can
    * be a tree ID (an ID that represents an entire AST) or it can be the ID
    * an individual AST node ID. Node IDs encode the tree ID and a portion that
    * is node-specific. The node-specific portion will be ignored.
    *
    * @param    id
    *           The ID to use as a key in the lookup. This ID may be a tree ID
    *           or a specific node ID.
    *
    * @return   The name associated with the tree or <code>null</code> if no
    *           such tree ID exists.
    */
   public String getTreeName(long id)
   {
      return plugin.getTreeName(id);
   }
   
   /**
    * Reverse lookup of the ID that corresponds to a given AST name.
    *
    * @param    name
    *           Tree name whose unique ID should be returned.
    *
    * @return   The tree ID associated with the name or
    *           {@link AstManager#INVALID_ID} if the tree name doesn't exist or
    *           if it is <code>null</code>.
    */
   public long getTreeId(String name)
   {
      return plugin.getTreeId(name);
   }
   
   /**
    * Returns the tree ID portion of a node ID.
    *
    * @return   The ID which corresponds to the tree itself, with all
    *           node ID portions removed (masked off) or
    *           {@link AstManager#INVALID_ID} if the ID is invalid.
    */
   public long getTreeId(long id)
   {
      return plugin.getTreeId(id);
   }
   
   /**
    * Load a named AST from persistent storage.
    *
    * @param    name
    *           The name of the tree. Often this is a file name.
    *
    * @return   The loaded AST.
    *
    * @throws   AstException
    *           On any error during the loading process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource doesn't exist.
    */
   public Aast loadTree(String name)
   throws AstException,
          IllegalArgumentException
   {
      return plugin.loadTree(name);
   }
   
   /**
    * Check if the AST specified by the given name exists already in the plugin's storage.
    * 
    * @param    name
    *           The name (including path) of the AST file to be checked.
    * 
    * @return   <code>true</code> if the specified AST is in the plugin's storage.
    */
   public boolean isExistingAst(String name)
   {
      return plugin.isExistingAst(name);
   }
   
   /**
    * Save the given AST to persistent storage under the given name.
    *
    * @param    ast
    *           The tree to save.
    * @param    name
    *           The name of the tree. Often this is a file name.
    * @param    terse
    *           Do not store line and column data when the node data is saved.
    * 
    * @throws   AstException
    *           On any error during the saving process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource cannot be
    *           used as a target to store the AST.
    */
   public void saveTree(Aast ast, String name, boolean terse)
   throws AstException,
          IllegalArgumentException
   {
      plugin.saveTree(ast, name, terse);
   }
   
   /**
    * Obtains a process-wide lock on any processing related to the given tree
    * name. If the lock is already held, this method will block until it
    * obtains the lock. When this method returns normally, the calling
    * thread is guaranteed to be holding the lock. This may be called
    * recursively by the same thread and if the current thread is already
    * holding the lock, then this method will only update the reference
    * count for the lock but otherwise it will immediately return. As a
    * consequence, the lock will not be released unless {@link #unlockTree}
    * is called the same number of times as this lock method. Thus the calls
    * to this lock method and to <code>unlock</code> must always be
    * balanced.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    name
    *           Tree name that is associated with the lock. This must be a
    *           valid tree name which can be used to uniquely identify the
    *           tree to be locked.
    *
    * @throws   InterruptedException
    *           When an interruption occurs during blocking for the lock. If
    *           this is thrown, the caller will NOT be holding the lock.
    * @throws   IllegalArgumentException
    *           If the name does not represent a valid tree.
    */
   public void lockTree(String name)
   throws InterruptedException,
          IllegalArgumentException
   {
      long id = getTreeId(name);
      
      if (id == INVALID_ID)
      {
         // tree doesn't yet exist, create it
         id = addTree(name);
      }
      
      lockTree(id);
   }
   
   /**
    * Obtains a process-wide lock on any processing related to the given tree.
    * If the lock is already held, this method will block until it obtains
    * the lock. When this method returns normally, the calling thread is
    * guaranteed to be holding the lock. This may be called recursively by the
    * same thread and if the current thread is already holding the lock, then
    * this method will only update the reference count for the lock but
    * otherwise it will immediately return. As a consequence, the lock will not
    * be released unless {@link #unlockTree} is called the same number of times
    * as this lock method. Thus the calls to this lock method and to
    * <code>unlock</code> must always be balanced.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    id
    *           Tree ID that is associated with the lock.
    *
    * @throws   InterruptedException
    *           When an interruption occurs during blocking for the lock. If
    *           this is thrown, the caller will NOT be holding the lock.
    * @throws   IllegalArgumentException
    *           If the ID is not a valid tree ID.
    */
   public void lockTree(long id)
   throws InterruptedException,
          IllegalArgumentException
   {
      plugin.lockTree(id);
   }
   
   /**
    * Releases the process-wide lock on any processing related to the given
    * tree. If the lock is already held by the calling thread, this method 
    * will decrement the reference count. If the resulting reference count is
    * 0, the lock will be released. If the calling thread does not hold the
    * lock, this method will exit abnormally. This may be called recursively by
    * the same thread and must be called the same number of times as the
    * {@link #lockTree} method. Thus the calls to this unlock method must always
    * be balanced with calls to this method.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    name
    *           Tree name that is associated with the lock. This must be a
    *           valid tree name which can be used to uniquely identify the
    *           tree to be locked.
    *
    * @throws   IllegalMonitorStateException
    *           If the calling thread does not own the lock.
    * @throws   IllegalArgumentException
    *           If the name does not represent a valid tree.
    */
   public void unlockTree(String name)
   throws IllegalMonitorStateException,
          IllegalArgumentException
   {
      unlockTree(getTreeId(name));
   }
   
   /**
    * Releases the process-wide lock on any processing related to the given
    * tree. If the lock is already held by the calling thread, this method 
    * will decrement the reference count. If the resulting reference count is
    * 0, the lock will be released. If the calling thread does not hold the
    * lock, this method will exit abnormally. This may be called recursively by
    * the same thread and must be called the same number of times as the
    * {@link #lockTree} method. Thus the calls to this unlock method must always
    * be balanced with calls to this method.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    id
    *           Tree ID that is associated with the lock.
    *
    * @throws   IllegalMonitorStateException
    *           If the calling thread does not own the lock.
    * @throws   IllegalArgumentException
    *           If the ID is not a valid tree ID.
    */
   public void unlockTree(long id)
   throws IllegalMonitorStateException,
          IllegalArgumentException
   {
      plugin.unlockTree(id);
   }

   /**
    * Access the next unassigned node ID for the specified tree. The
    * value is looked up and returned, just before returning the internal
    * cache is incremented and stored back for next use. Once this value is
    * returned, it is considered "used" and will not be returned again.
    *
    * @param    id
    *           The ID that specifies a tree.
    *
    * @return   The newly assigned node ID or {@link #INVALID_ID} if no such
    *           tree exists.
    *
    * @throws   AstException
    *           If too many AST nodes exist.
    */
   public long getNextNodeId(long id)
   throws AstException
   {
      return plugin.getNextNodeId(id);
   }

   /**
    * Resets the next unassigned AST node ID for the specified tree to the
    * first possible value (tree ID + 1).
    * <p>
    * <b>WARNING: there is no check on the current usage of node IDs for this
    * file, so if AST nodes exist already they will have IDs at or above this
    * value and duplicate IDs may (probably will) result!</b>
    *
    * @param    treeId
    *           The ID that specifies a valid tree.
    *
    * @throws   AstException
    *           If the file ID is invalid.
    */
   public void resetNextNodeId(long treeId)
   throws AstException
   {
      plugin.resetNextNodeId(treeId);
   }

   /**
    * Adds a mapping from the specified AST to a new ID. The next available
    * tree ID is assigned and this mapping is added to the registry. If the
    * tree name already exists, the existing tree ID is returned.
    *
    * @param    name
    *           The tree name. Often this will be the name of a file.
    *
    * @return   The assigned tree ID.
    *
    * @throws   AstException
    *           If too many ASTs exist in the registry or if the input
    *           name is <code>null</code>.
    */
   public long addTree(String name)
   throws AstException
   {
      return plugin.addTree(name);
   }

   /**
    * Removes all mapping data referencing the specified tree.
    *
    * @param    id
    *           The ID the tree to remove.
    *
    * @return   <code>true</code> if the tree existed in the registry (and 
    *           was removed), <code>false</code> if the tree didn't exist.
    */
   public boolean removeTree(long id)
   throws AstException
   {
      return plugin.removeTree(id);
   }
   
   /**
    * Removes all mapping data referencing the specified tree.
    *
    * @param    name
    *           The tree name to remove.
    *
    * @return   <code>true</code> if the tree existed in the registry (and 
    *           was removed), <code>false</code> if the tree didn't exist.
    */
   public boolean removeTree(String name)
   throws AstException
   {
      return plugin.removeTree(getTreeId(name));
   }
   
   /**
    * Persistently store the current state of the AST manager plugin.
    *
    * @throws  AstException
    *          On any failure.
    */
   public void save()
   throws AstException
   {
      plugin.save();
   }
}
