/*
** Module   : DumpTree.java
** Abstract : print an Aast to a user-defined output stream or to stdout 
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20041117   @18671 First version based on antlr.DumpASTVisitor.
**                           The formatting was changed to meet GCD
**                           coding standards.  Most importantly, it
**                           now prints token type names instead of the
**                           integer token type itself. Moved the package
**                           to include it in util.  Added javadoc.
** 002 ECF 20050124   @19406 Added option to report position (line and
**                           column numbers) at which underlying token was
**                           found.
** 003 GES 20050126   @19408 Added support for user control over the
**                           output stream.  This allows the caller to
**                           redirect the output to a file or other
**                           destination.
** 004 GES 20050303   @20163 Moved from util to uast package and switched
**                           to using the Aast interface for all dump
**                           activities. Rewrote the code to fix a defect
**                           in the visitor which would walk right side
**                           siblings of the node passed in.  This worked
**                           in the past because we only used it on a
**                           root node which has no such siblings, so the
**                           bug was not seen. There is no code left from
**                           the original antlr.DumpASTVisitor
**                           implementation, it is all GCD now.
** 005 EVL 20070609   @34005 Renaming reserved in Java 6 enum identifier
**                           to be able to compile for Java 6.
** 006 GES 20090424   @41954 Import change.
** 007 GES 20090427   @42017 Moved this class to a new package.
** 008 GES 20090518   @42371 Import change.
** 009 CA  20131023          Added debug aid to dump annotation info.
*/

package com.goldencode.ast; 
 
import java.io.*;
import java.util.*;
import com.goldencode.util.*;

/**
 * Dumps the contents of an <code>AST</code> to <code>stdout</code> or to
 * a user-defined output stream. 
 * <p>
 * Each node prints the symbolic text name of the token type rather than the
 * integer representation. The token text, line and column number information
 * for the AST will also be rendered.
 */
public class DumpTree
{
   /** Debug aid to dump annotations information. */
   private static final boolean DUMP_ANNOTATIONS = false;
   
   /** The destination of all dumped output. */
   private PrintStream out = null;
   
   /**
    * Default constructor which uses <code>stdout</code>.
    */
   public DumpTree()
   {
      this(null);
   }   
   
   /**
    * Constructs an instance and allows control over the output destination.
    * If the output stream is <code>null</code> it will be set to 
    * <code>System.out</code>.
    *
    * @param    out
    *           The new stream to use as the output destination.
    */
   public DumpTree(PrintStream out)
   {
      setOutputStream(out);
   }   
   
   /**
    * Allows the current output stream to be switched for a user-defined
    * stream.  If set to <code>null</code>, <code>System.out</code> will
    * be used.  The caller must close this stream when done!
    *
    * @param    out
    *           The new stream to use as the output destination.
    */
   public void setOutputStream(PrintStream out)
   {
      if (out != null)
      {
         this.out = out;
      }
      else
      {
         this.out = System.out;
      }      
   }
   
   /**
    * Prints a 3 space indent for every additional nested level in the tree.
    */
   private void printIndent(int tabs)
   {
      if (tabs > 0)
      {
         out.print(StringHelper.repeatChar(' ', tabs*3));
      }
   }

   /**
    * Prints out a text-based representation of an <code>Aast</code> and
    * all children.  This method relies upon the fact that all tree nodes 
    * represent a full tree themselves, even if they have no children. An
    * iterator is obtained that allows a simple depth-first, left-right
    * walk of the tree in a single loop.
    * 
    * @param   node
    *          The root of the tree or subtree which should be printed.
    */
   public void visit(Aast node)
   {
      Iterator iter = node.iterator();
      int      root = node.getDepth();
      
      while (iter.hasNext())
      {
         Aast next = (Aast) iter.next();
         
         printIndent(next.getDepth() - root);
         
         StringBuffer sb = new StringBuffer(next.getText());
         
         // print the token type as a name
         sb.append(" [").append(next.getSymbolicTokenType()).append("] ");

         // print token position
         sb.append("@").append(next.getLine()).append(':');
         sb.append(next.getColumn());
            
         out.println(sb.toString());
         
         // output annotations (only for debugging)
         Iterator<String> itra = next.annotationKeys();
         if (DUMP_ANNOTATIONS && itra.hasNext())
         {
            printIndent(next.getDepth() - root);
            sb = new StringBuffer();
            sb.append("(");
            while (itra.hasNext())
            {
               String key = itra.next();
               sb.append(key).append("=").append(next.getAnnotation(key));
               if (itra.hasNext())
               {
                  sb.append(", ");
               }
            }
            sb.append(")");
            
            out.println(sb.toString());
         }
      }      
   }
}
